require 'csv'
require 'date'
require 'haml'
require 'json'
require 'sinatra/base'
require 'sinatra/flash'


module Graffable

  class App < Sinatra::Base

    configure do
      set      :root,   File.dirname(__FILE__)
      set      :haml,   { format: :html5 }
      set      :static, true
      enable   :sessions

      register Sinatra::Flash
    end


    before do
      @db     = Graffable::Database.connect
      @groups = @db[:groups].order(:name)
    end


    # /:group/:report/:year/:month/:day.json
    get %r{/(.+)/(.+)/(\d{4})/(\d{2})/(\d{2}).(csv|json)} do |group_name, report_name, year, month, day, extension|
      group  = assert_group  group_name
      report = assert_report group, report_name

      # TODO DRY!
      query   = "SELECT year,month,day,hour,label,#{ report[:aggregate].to_s.upcase }(value) AS value "  +
                "FROM numbers "                                                                          +
                "WHERE report_id=:report_id "                                                            +
                "AND year=:year "                                                                        +
                "AND month=:month "                                                                      +
                "AND day=:day "                                                                          +
                "GROUP BY year,month,day,hour,label "                                                    +
                "ORDER BY year,month,day,hour,label"
      dataset = @db.fetch( query, day: day, month: month, report_id: report[:id], year: year )

      case extension
      when 'csv'
        return_csv dataset
      when 'json'
        return_json dataset, date_formatter: lambda { |y,m,d,h| h }
      end
    end

    # /:group/:report/:year/:month.json
    get %r{/(.+)/(.+)/(\d{4})/(\d{2}).(csv|json)} do |group_name, report_name, year, month, extension|
      group  = assert_group  group_name
      report = assert_report group, report_name

      # TODO DRY!
      query   = "SELECT year,month,day,label,#{ report[:aggregate].to_s.upcase }(value) AS value "  +
                "FROM numbers "                                                                     +
                "WHERE report_id=:report_id "                                                       +
                "AND year=:year "                                                                   +
                "AND month=:month "                                                                 +
                "GROUP BY year,month,day,label "                                                    +
                "ORDER BY year,month,day,label"
      dataset = @db.fetch( query, month: month, report_id: report[:id], year: year )

      case extension
      when 'csv'
        return_csv dataset
      when 'json'
        return_json dataset, date_formatter: lambda { |y,m,d,h| d }
      end
    end

    # /:group/:report/:year.json
    get %r{/(.+)/(.+)/(\d{4}).(csv|json)} do |group_name, report_name, year, extension|
      group  = assert_group  group_name
      report = assert_report group, report_name

      # TODO DRY!
      query   = "SELECT year,month,label,#{ report[:aggregate].to_s.upcase }(value) AS value "  +
                "FROM numbers "                                                                 +
                "WHERE report_id=:report_id "                                                   +
                "AND year=:year "                                                               +
                "GROUP BY year,month,label "                                                    +
                "ORDER BY year,month,label"
      dataset = @db.fetch( query, report_id: report[:id], year: year )

      case extension
      when 'csv'
        return_csv dataset
      when 'json'
        return_json dataset, date_formatter: lambda { |y,m,d,h| "#{y}-#{m}" }
      end
    end

    # /:group/:report.json
    get %r{/(.+)/(.+).(csv|json)} do |group_name, report_name, extension|
      group  = assert_group  group_name
      report = assert_report group, report_name

      # TODO DRY!
      query   = "SELECT year,month,day,label,#{ report[:aggregate].to_s.upcase }(value) AS value "  +
                "FROM numbers "                                                                     +
                "WHERE report_id=:report_id "                                                       +
                "GROUP BY year,month,day,label "                                                    +
                "ORDER BY year,month,day,label"
      dataset = @db.fetch( query, report_id: report[:id] )
      max     = 15

      case extension
      when 'csv'
        return_csv dataset, max: max
      when 'json'
        return_json dataset, date_formatter: lambda { |y,m,d,h| Date.parse("#{y}-#{m}-#{d}").strftime("%b %d") }, max: max
      end
    end

    # TODO DRY!
    get %r{/(.+)/(.+)/(\d{4})/(\d{2})/(\d{2})} do |group_name, report_name, year, month, day|
      @group = @groups.detect { |g| group_name == g[:name] }
      redirect_with_warning( '/', "Report group not found: #{group_name}" ) unless @group

      @reports = @db[:reports].where( group_id: @group[:id] ).order(:description).all
      @report  = @reports.detect { |r| report_name == r[:name] }
      redirect_with_warning( "/#{group_name}", "Report not found: #{report_name}" ) unless @report

      @data_url = "/#{group_name}/#{report_name}/#{year}/#{month}/#{day}.json"
      @interval = "#{year}-#{month}-#{day}"
      
      date      = Date.parse( "#{year}-#{month}-#{day}" )
      prev_day  = date - 1
      next_day  = date + 1
      @previous = "/#{group_name}/#{report_name}/#{ prev_day.year }/#{ prev_day.strftime('%m') }/#{ prev_day.strftime('%d') }"
      @next     = "/#{group_name}/#{report_name}/#{ next_day.year }/#{ next_day.strftime('%m') }/#{ next_day.strftime('%d') }"

      @title = "#{ @report[:description] } - #{ @interval }"

      haml :report
    end

    # TODO DRY!
    get %r{/(.+)/(.+)/(\d{4})/(\d{2})} do |group_name, report_name, year, month|
      @group = @groups.detect { |g| group_name == g[:name] }
      redirect_with_warning( '/', "Report group not found: #{group_name}" ) unless @group

      @reports = @db[:reports].where( group_id: @group[:id] ).order(:description).all
      @report  = @reports.detect { |r| report_name == r[:name] }
      redirect_with_warning( "/#{group_name}", "Report not found: #{report_name}" ) unless @report

      @data_url = "/#{group_name}/#{report_name}/#{year}/#{month}.json"
      @interval = "#{year}-#{month}"
      
      date      = Date.parse( "#{year}-#{month}-01" )
      @previous = "/#{group_name}/#{report_name}/#{ date.prev_month.year }/#{ date.prev_month.strftime('%m') }"
      @next     = "/#{group_name}/#{report_name}/#{ date.next_month.year }/#{ date.next_month.strftime('%m') }"

      @title = "#{ @report[:description] } - #{ @interval }"

      haml :report
    end

    # TODO DRY!
    get %r{/(.+)/(.+)/(\d{4})} do |group_name, report_name, year|
      @group = @groups.detect { |g| group_name == g[:name] }
      redirect_with_warning( '/', "Report group not found: #{group_name}" ) unless @group

      @reports = @db[:reports].where( group_id: @group[:id] ).order(:description).all 
      @report  = @reports.detect { |r| report_name == r[:name] }
      redirect_with_warning( "/#{group_name}", "Report not found: #{report_name}" ) unless @report

      @data_url = "/#{group_name}/#{report_name}/#{year}.json"
      @interval = year

      date      = Date.parse( "#{year}-01-01" )
      @previous = "/#{group_name}/#{report_name}/#{ date.prev_year.year }"
      @next     = "/#{group_name}/#{report_name}/#{ date.next_year.year }"

      @title = "#{ @report[:description] } - #{ @interval }"

      haml :report
    end

    # TODO DRY!
    get '/:group/:report' do
      @group = @groups.detect { |g| params[:group] == g[:name] }
      redirect_with_warning( '/', "Report group not found: #{ params[:group] }" ) unless @group

      @reports = @db[:reports].where( group_id: @group[:id] ).order(:description).all 
      @report  = @reports.detect { |r| params[:report] == r[:name] }
      redirect_with_warning( "/#{ params[:group] }", "Report not found: #{ params[:report] }" ) unless @report

      @data_url = "/#{ params[:group] }/#{ params[:report] }.json"
      @interval = 'Recent'

      @title = "#{ @report[:description] } - #{ @interval }"

      haml :report
    end

    get '/:group' do
      @group   = @groups.detect{ |g| params[:group] == g[:name] }
      redirect_with_warning( '/', "Report group not found: #{ params[:group] }" ) unless @group

      @reports = @db[:reports].where( group_id: @group[:id] ).order(:description).all
      haml :group
    end

    get '/' do
      haml :index
    end


    private

      def assert_group(group_name)
        group = @groups.detect { |g| group_name == g[:name] }
        halt 404, "Report group not found: #{group_name}" unless group
        group
      end

      def assert_report(group, report_name)
        reports = @db[:reports].where( group_id: group[:id] ).all 
        report  = reports.detect { |r| report_name == r[:name] }
        halt 404, "Report not found: #{report_name}" unless report
        report
      end

      def redirect_with_warning(path, message)
        flash[:warning] = message
        redirect to(path)
      end

      def return_csv( dataset, params = {} )
        defaults = { max: -1 }
        opts     = defaults.merge params

        data   = [ %w( date label value ).to_csv ]
        values = {}
        dataset.each do |row|
          date   = %i( year month day hour ).collect { |k| row[k] }.compact.join('-')
          label  = row[:label] || ''
          value  = row[:value].to_i

          values[date]        ||= {}
          values[date][label]   = value
        end

        values = values.sort.reverse.slice( 0 .. opts[:max] - 1 ).reverse if opts[:max] > - 1
        values.each do |tuple|
          date  = tuple.first
          pairs = tuple.last
          pairs.each_pair do |label, value|
            data << [ date, label, value ].to_csv
          end
        end

        content_type :text
        return data.join
      end

      def return_json( dataset, params = {} )
        defaults = { max: -1 }
        opts     = defaults.merge params

        data         = {}
        date_mapping = {}
        values       = {}
        dataset.each do |row|
          date  = %i( year month day hour ).collect { |k| row[k] }.compact.join('-')
          label = row[:label] || ''
          value = row[:value].to_i
        
          date_mapping[date] = opts[:date_formatter].call row[:year], row[:month], row[:day], row[:hour] unless date_mapping.key?(date)

          values[date]        ||= {}
          values[date][label]   = value
        end

        values = values.sort.reverse.slice( 0 .. opts[:max] - 1 ).reverse if opts[:max] > - 1
        values.each do |tuple|
          date  = date_mapping[ tuple.first ]
          pairs = tuple.last
          pairs.each_pair do |label, value|
            unless data.key?(label)
              data[label]         = { data: [], total: 0 }
              data[label][:label] = label unless label.empty?
            end
            data[label][:data].push [ date, value ]
            data[label][:total] = data[label][:total] + value
            data[label][:label] = "#{label} (#{ data[label][:total] })" unless label.empty?
          end
        end

        content_type :json
        return { data: data.values }.to_json
      end

  end # class Graffable::App
end   # module Graffable

